@@ -7,10 +7,13 @@ use idmap::{GidMap, IdMapSetUpPipeMessage, UidMap};
 use std::ffi::CString;
 use std::fs::{self, File};
 use std::io::{Read, Write};
+use std::os::fd::OwnedFd;
 use std::os::unix::io::{AsRawFd, FromRawFd};
+use std::path::Path;
 use std::process::{self, Command};
 use std::str::FromStr;
 use std::{error, fmt, io};
+use vhost::vhost_user::Listener;
 
 #[derive(Debug)]
 pub enum Error {
@@ -152,17 +155,17 @@ pub struct Sandbox {
     /// Mechanism to be used for setting up the sandbox.
     sandbox_mode: SandboxMode,
     /// UidMap to be used for `newuidmap(1)` command line arguments
-    uid_map: Option<UidMap>,
+    uid_map: Vec<UidMap>,
     /// GidMap to be used for `newgidmap(1)` command line arguments
-    gid_map: Option<GidMap>,
+    gid_map: Vec<GidMap>,
 }
 
 impl Sandbox {
     pub fn new(
         shared_dir: String,
         sandbox_mode: SandboxMode,
-        uid_map: Option<UidMap>,
-        gid_map: Option<GidMap>,
+        uid_map: Vec<UidMap>,
+        gid_map: Vec<GidMap>,
     ) -> io::Result<Self> {
         let shared_dir_rp = fs::canonicalize(shared_dir)?;
         let shared_dir_rp_str = shared_dir_rp
@@ -310,66 +313,102 @@ impl Sandbox {
     /// Sets mappings for the given uid and gid.
     fn setup_id_mappings(
         &self,
-        uid_map: Option<&UidMap>,
-        gid_map: Option<&GidMap>,
+        uid_map: &[UidMap],
+        gid_map: &[GidMap],
         pid: i32,
     ) -> Result<(), Error> {
+        let current_uid = unsafe { libc::geteuid() };
+        let current_gid = unsafe { libc::getegid() };
+
+        // Take uid map or set up a 1-to-1 mapping for our current euid.
+        let default_uid_map = vec![UidMap {
+            outside_uid: current_uid,
+            inside_uid: current_uid,
+            count: 1,
+        }];
+        let uid_map = if uid_map.is_empty() {
+            &default_uid_map
+        } else {
+            uid_map
+        };
+
+        // Take gid map or set up a 1-to-1 mapping for our current gid.
+        let default_gid_map = vec![GidMap {
+            outside_gid: current_gid,
+            inside_gid: current_gid,
+            count: 1,
+        }];
+        let gid_map = if gid_map.is_empty() {
+            &default_gid_map
+        } else {
+            gid_map
+        };
+
         // Unprivileged user can not set any mapping without any restriction.
         // Therefore, newuidmap/newgidmap is used instead of writing directly
-        // into proc/[pid]/{uid,gid}_map.
-        let mut newuidmap = Command::new("newuidmap");
-        newuidmap.arg(pid.to_string());
-        if let Some(uid_map) = uid_map {
-            newuidmap.arg(uid_map.inside_uid.to_string());
-            newuidmap.arg(uid_map.outside_uid.to_string());
-            newuidmap.arg(uid_map.count.to_string());
+        // into proc/[pid]/{uid,gid}_map if a potentially privileged action is
+        // requested (outside {u,g}id != e{u,g}id or count > 1).
+        if uid_map.len() != 1 || uid_map[0].outside_uid != current_uid || uid_map[0].count > 1 {
+            let mut newuidmap = Command::new("newuidmap");
+            newuidmap.arg(pid.to_string());
+            for entry in uid_map.iter() {
+                newuidmap.arg(entry.inside_uid.to_string());
+                newuidmap.arg(entry.outside_uid.to_string());
+                newuidmap.arg(entry.count.to_string());
+            }
+            let output = newuidmap.output().map_err(|_| {
+                Error::WriteUidMap(format!(
+                    "failed to execute newuidmap: {}",
+                    io::Error::last_os_error()
+                ))
+            })?;
+            if !output.status.success() {
+                return Err(Error::WriteUidMap(
+                    String::from_utf8_lossy(&output.stderr).to_string(),
+                ));
+            }
         } else {
-            // Set up 1-to-1 mappings for our current uid.
-            let current_uid = unsafe { libc::geteuid() };
-            newuidmap.arg(current_uid.to_string());
-            newuidmap.arg(current_uid.to_string());
-            newuidmap.arg("1");
-        }
-        let mut output = newuidmap.output().map_err(|_| {
-            Error::WriteUidMap(format!(
-                "failed to execute newuidmap: {}",
-                io::Error::last_os_error()
-            ))
-        })?;
-        if !output.status.success() {
-            return Err(Error::WriteUidMap(
-                String::from_utf8_lossy(&output.stderr).to_string(),
-            ));
+            // Unprivileged part, we can driectly write to /proc/[pid]/uid_map.
+            std::fs::write(
+                format!("/proc/{pid}/uid_map"),
+                format!("{} {} 1", uid_map[0].inside_uid, uid_map[0].outside_uid),
+            )
+            .map_err(|e| Error::WriteUidMap(e.to_string()))?;
         }
 
-        let mut newgidmap = Command::new("newgidmap");
-        newgidmap.arg(pid.to_string());
-        if let Some(gid_map) = gid_map {
-            newgidmap.arg(gid_map.inside_gid.to_string());
-            newgidmap.arg(gid_map.outside_gid.to_string());
-            newgidmap.arg(gid_map.count.to_string());
+        if gid_map.len() != 1 || gid_map[0].outside_gid != current_gid || gid_map[0].count > 1 {
+            let mut newgidmap = Command::new("newgidmap");
+            newgidmap.arg(pid.to_string());
+            for entry in gid_map.iter() {
+                newgidmap.arg(entry.inside_gid.to_string());
+                newgidmap.arg(entry.outside_gid.to_string());
+                newgidmap.arg(entry.count.to_string());
+            }
+            let output = newgidmap.output().map_err(|_| {
+                Error::WriteGidMap(format!(
+                    "failed to execute newgidmap: {}",
+                    io::Error::last_os_error()
+                ))
+            })?;
+            if !output.status.success() {
+                return Err(Error::WriteGidMap(
+                    String::from_utf8_lossy(&output.stderr).to_string(),
+                ));
+            }
         } else {
-            // Set up 1-to-1 mappings for our current gid.
-            let current_gid = unsafe { libc::getegid() };
-            newgidmap.arg(current_gid.to_string());
-            newgidmap.arg(current_gid.to_string());
-            newgidmap.arg("1");
-        }
-        output = newgidmap.output().map_err(|_| {
-            Error::WriteGidMap(format!(
-                "failed to execute newgidmap: {}",
-                io::Error::last_os_error()
-            ))
-        })?;
-        if !output.status.success() {
-            return Err(Error::WriteGidMap(
-                String::from_utf8_lossy(&output.stderr).to_string(),
-            ));
+            // Unprivileged part, we can driectly write to /proc/[pid]/gid_map.
+            std::fs::write(format!("/proc/{pid}/setgroups"), b"deny")
+                .map_err(|e| Error::WriteGidMap(e.to_string()))?;
+            std::fs::write(
+                format!("/proc/{pid}/gid_map"),
+                format!("{} {} 1", gid_map[0].inside_gid, gid_map[0].outside_gid),
+            )
+            .map_err(|e| Error::WriteGidMap(e.to_string()))?;
         }
         Ok(())
     }
 
-    pub fn enter_namespace(&mut self) -> Result<(), Error> {
+    pub fn enter_namespace(&mut self, listener: Listener) -> Result<Listener, Error> {
         let uid = unsafe { libc::geteuid() };
 
         let flags = if uid == 0 {
@@ -402,9 +441,7 @@ impl Sandbox {
             // Setup uid/gid mappings
             if uid != 0 {
                 let ppid = unsafe { libc::getppid() };
-                if let Err(error) =
-                    self.setup_id_mappings(self.uid_map.as_ref(), self.gid_map.as_ref(), ppid)
-                {
+                if let Err(error) = self.setup_id_mappings(&self.uid_map, &self.gid_map, ppid) {
                     // We don't really need to close the pipes here, since the OS will close the FDs
                     // after the process exits. But let's do it explicitly to signal an error to the
                     // other end of the pipe.
@@ -463,9 +500,27 @@ impl Sandbox {
             if child == 0 {
                 // Second child
                 self.setup_mounts()?;
-                Ok(())
+                Ok(listener)
             } else {
                 // This is the parent
+
+                // The child process drops the `vhost::Listener` after the first
+                // `accept()`. However, since the parent just waits until the child
+                // ends, keeping all the FDs open, as well as the socket's FD in a
+                // listen state. This is problematic because nothing prevents a
+                // miss-configured VMM to try to connect twice to the same socket
+                // leaving the VMM waiting forever. So, let's close the listener
+                // before waiting for the child.
+                let fd = listener.as_raw_fd();
+
+                // `vhost::Listener` beside closing the FD, it will remove the socket, if dropped
+                std::mem::forget(listener);
+
+                // Let's close the FD without removing the socket file
+                // SAFETY: `fd` is open and nobody owns it
+                let fd = unsafe { OwnedFd::from_raw_fd(fd) };
+                drop(fd);
+
                 util::wait_for_child(child); // This never returns.
             }
         }
@@ -509,6 +564,12 @@ impl Sandbox {
             return Ok(false);
         }
 
+        // If we are running as root and the system does not support user namespaces,
+        // we must drop supplemental groups.
+        if !Path::new("/proc/self/ns/user").exists() {
+            return Ok(true);
+        }
+
         let uid_mmap_data =
             fs::read_to_string("/proc/self/uid_map").map_err(Error::DropSupplementalGroups)?;
         let uid_map: Vec<_> = uid_mmap_data.split_whitespace().collect();
@@ -545,17 +606,17 @@ impl Sandbox {
     }
 
     /// Set up sandbox,
-    pub fn enter(&mut self) -> Result<(), Error> {
+    pub fn enter(&mut self, listener: Listener) -> Result<Listener, Error> {
         let uid = unsafe { libc::geteuid() };
         if uid != 0 && self.sandbox_mode == SandboxMode::Chroot {
             return Err(Error::SandboxModeInvalidUID);
         }
 
-        if self.uid_map.is_some() && (uid == 0 || self.sandbox_mode != SandboxMode::Namespace) {
+        if !self.uid_map.is_empty() && (uid == 0 || self.sandbox_mode != SandboxMode::Namespace) {
             return Err(Error::SandboxModeInvalidUidMap);
         }
 
-        if self.gid_map.is_some() && (uid == 0 || self.sandbox_mode != SandboxMode::Namespace) {
+        if !self.gid_map.is_empty() && (uid == 0 || self.sandbox_mode != SandboxMode::Namespace) {
             return Err(Error::SandboxModeInvalidGidMap);
         }
 
@@ -567,14 +628,25 @@ impl Sandbox {
         // possible and thus there's no need to drop supplemental groups. In
         // both of these scenarios calling setgroups() is also not allowed so we
         // avoid calling it since we know it will return a privilege error.
-        if self.must_drop_supplemental_groups()? {
+        let must_drop_supplemental_groups = match self.must_drop_supplemental_groups() {
+            Ok(must_drop) => must_drop,
+            Err(error) => {
+                warn!(
+                    "Failed to determine whether supplemental groups must be dropped: {error}; \
+                    defaulting to trying to drop supplemental groups"
+                );
+                true
+            }
+        };
+
+        if must_drop_supplemental_groups {
             self.drop_supplemental_groups()?;
         }
 
         match self.sandbox_mode {
-            SandboxMode::Namespace => self.enter_namespace(),
-            SandboxMode::Chroot => self.enter_chroot(),
-            SandboxMode::None => Ok(()),
+            SandboxMode::Namespace => self.enter_namespace(listener),
+            SandboxMode::Chroot => self.enter_chroot().and(Ok(listener)),
+            SandboxMode::None => Ok(listener),
         }
     }
 
