@@ -2,21 +2,22 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-use super::fs_cache_req_handler::FsCacheReqHandler;
 use crate::descriptor_utils::{Reader, Writer};
 use crate::filesystem::{
     Context, DirEntry, DirectoryIterator, Entry, Extensions, FileSystem, GetxattrReply,
-    ListxattrReply, SecContext, ZeroCopyReader, ZeroCopyWriter,
+    ListxattrReply, SecContext, SerializableFileSystem, ZeroCopyReader, ZeroCopyWriter,
 };
 use crate::fuse::*;
 use crate::passthrough::util::einval;
+use crate::soft_idmap::GuestGid;
 use crate::{oslib, Error, Result};
 use std::convert::{TryFrom, TryInto};
 use std::ffi::{CStr, CString};
 use std::fs::File;
 use std::io::{self, Read, Write};
-use std::mem::{size_of, MaybeUninit};
-use std::sync::atomic::{AtomicU64, Ordering};
+use std::mem::size_of;
+use std::sync::atomic::{AtomicBool, AtomicU64, Ordering};
+use std::sync::Arc;
 use std::time::Duration;
 use vm_memory::ByteValued;
 
@@ -29,39 +30,53 @@ const PARENT_DIR_CSTR: &[u8] = b"..";
 
 struct ZcReader<'a>(Reader<'a>);
 
-impl<'a> ZeroCopyReader for ZcReader<'a> {
-    fn read_to(
+impl ZeroCopyReader for ZcReader<'_> {
+    fn write_to_file_at(
         &mut self,
         f: &File,
-        count: usize,
-        off: u64,
+        mut count: usize,
+        mut off: u64,
         flags: Option<oslib::WritevFlags>,
     ) -> io::Result<usize> {
-        self.0.read_to_at(f, count, off, flags)
-    }
-}
+        let start = off;
+        while count > 0 {
+            let written = self.0.write_to_file_at(f, count, off, flags)?;
+            if written == 0 {
+                break;
+            }
+            off = off
+                .checked_add(written as u64)
+                .ok_or_else(|| io::Error::other("Write wrap-around"))?;
+            count = count
+                .checked_sub(written)
+                .ok_or_else(|| io::Error::other("Write operation wrote more than requested"))?;
+        }
 
-impl<'a> io::Read for ZcReader<'a> {
-    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
-        self.0.read(buf)
+        // Must fit: Cannot be greater than `count` originally
+        Ok((off - start) as usize)
     }
 }
 
 struct ZcWriter<'a>(Writer<'a>);
 
-impl<'a> ZeroCopyWriter for ZcWriter<'a> {
-    fn write_from(&mut self, f: &File, count: usize, off: u64) -> io::Result<usize> {
-        self.0.write_from_at(f, count, off)
-    }
-}
+impl ZeroCopyWriter for ZcWriter<'_> {
+    fn read_from_file_at(&mut self, f: &File, mut count: usize, mut off: u64) -> io::Result<usize> {
+        let start = off;
+        while count > 0 {
+            let read = self.0.read_from_file_at(f, count, off)?;
+            if read == 0 {
+                break;
+            }
+            off = off
+                .checked_add(read as u64)
+                .ok_or_else(|| io::Error::other("Read wrap-around"))?;
+            count = count
+                .checked_sub(read)
+                .ok_or_else(|| io::Error::other("Read operation wrote more than requested"))?;
+        }
 
-impl<'a> io::Write for ZcWriter<'a> {
-    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
-        self.0.write(buf)
-    }
-
-    fn flush(&mut self) -> io::Result<()> {
-        self.0.flush()
+        // Must fit: Cannot be greater than `count` originally
+        Ok((off - start) as usize)
     }
 }
 
@@ -79,7 +94,7 @@ impl<F: FileSystem + Sync> Server<F> {
     }
 
     #[allow(clippy::cognitive_complexity)]
-    pub fn handle_message<T: FsCacheReqHandler>(
+    pub fn handle_message<T>(
         &self,
         mut r: Reader,
         w: Writer,
@@ -164,92 +179,32 @@ impl<F: FileSystem + Sync> Server<F> {
         }
     }
 
-    fn setupmapping<T: FsCacheReqHandler>(
+    fn setupmapping<T>(
         &self,
         in_header: InHeader,
-        mut r: Reader,
+        _r: Reader,
         w: Writer,
-        vu_req: Option<&mut T>,
+        _vu_req: Option<&mut T>,
     ) -> Result<usize> {
-        if let Some(req) = vu_req {
-            let SetupmappingIn {
-                fh,
-                foffset,
-                len,
-                flags,
-                moffset,
-            } = r.read_obj().map_err(Error::DecodeMessage)?;
-
-            match self.fs.setupmapping(
-                Context::from(in_header),
-                in_header.nodeid.into(),
-                fh.into(),
-                foffset,
-                len,
-                flags,
-                moffset,
-                req,
-            ) {
-                Ok(()) => reply_ok(None::<u8>, None, in_header.unique, w),
-                Err(e) => reply_error(e, in_header.unique, w),
-            }
-        } else {
-            reply_error(
-                io::Error::from_raw_os_error(libc::EINVAL),
-                in_header.unique,
-                w,
-            )
-        }
+        reply_error(
+            io::Error::from_raw_os_error(libc::ENOSYS),
+            in_header.unique,
+            w,
+        )
     }
 
-    fn removemapping<T: FsCacheReqHandler>(
+    fn removemapping<T>(
         &self,
         in_header: InHeader,
-        mut r: Reader,
+        _r: Reader,
         w: Writer,
-        vu_req: Option<&mut T>,
+        _vu_req: Option<&mut T>,
     ) -> Result<usize> {
-        if let Some(req) = vu_req {
-            let RemovemappingIn { count } = r.read_obj().map_err(Error::DecodeMessage)?;
-
-            if let Some(size) = (count as usize).checked_mul(size_of::<RemovemappingOne>()) {
-                if size > MAX_BUFFER_SIZE as usize {
-                    return reply_error(
-                        io::Error::from_raw_os_error(libc::ENOMEM),
-                        in_header.unique,
-                        w,
-                    );
-                }
-            } else {
-                return reply_error(
-                    io::Error::from_raw_os_error(libc::EOVERFLOW),
-                    in_header.unique,
-                    w,
-                );
-            }
-
-            let mut requests = Vec::with_capacity(count as usize);
-            for _ in 0..count {
-                requests.push(
-                    r.read_obj::<RemovemappingOne>()
-                        .map_err(Error::DecodeMessage)?,
-                );
-            }
-
-            match self
-                .fs
-                .removemapping(Context::from(in_header), requests, req)
-            {
-                Ok(()) => reply_ok(None::<u8>, None, in_header.unique, w),
-                Err(e) => reply_error(e, in_header.unique, w),
-            }
-        } else {
-            reply_error(
-                io::Error::from_raw_os_error(libc::EINVAL),
-                in_header.unique,
-                w,
-            )
-        }
+        reply_error(
+            io::Error::from_raw_os_error(libc::ENOSYS),
+            in_header.unique,
+            w,
+        )
     }
 
     fn lookup(&self, in_header: InHeader, mut r: Reader, w: Writer) -> Result<usize> {
@@ -299,12 +254,12 @@ impl<F: FileSystem + Sync> Server<F> {
             .fs
             .getattr(Context::from(in_header), in_header.nodeid.into(), handle)
         {
-            Ok((st, timeout)) => {
+            Ok((attr, timeout)) => {
                 let out = AttrOut {
                     attr_valid: timeout.as_secs(),
                     attr_valid_nsec: timeout.subsec_nanos(),
                     dummy: 0,
-                    attr: st.into(),
+                    attr,
                 };
                 reply_ok(Some(out), None, in_header.unique, w)
             }
@@ -323,21 +278,19 @@ impl<F: FileSystem + Sync> Server<F> {
 
         let valid = SetattrValid::from_bits_truncate(setattr_in.valid);
 
-        let st: libc::stat64 = setattr_in.into();
-
         match self.fs.setattr(
             Context::from(in_header),
             in_header.nodeid.into(),
-            st,
+            setattr_in,
             handle,
             valid,
         ) {
-            Ok((st, timeout)) => {
+            Ok((attr, timeout)) => {
                 let out = AttrOut {
                     attr_valid: timeout.as_secs(),
                     attr_valid_nsec: timeout.subsec_nanos(),
                     dummy: 0,
-                    attr: st.into(),
+                    attr,
                 };
                 reply_ok(Some(out), None, in_header.unique, w)
             }
@@ -978,7 +931,8 @@ impl<F: FileSystem + Sync> Server<F> {
             | FsOptions::MAX_PAGES
             | FsOptions::SUBMOUNTS
             | FsOptions::INIT_EXT
-            | FsOptions::CREATE_SUPP_GROUP;
+            | FsOptions::CREATE_SUPP_GROUP
+            | FsOptions::ALLOW_IDMAP;
 
         let flags_64 = ((flags2 as u64) << 32) | (flags as u64);
         let capable = FsOptions::from_bits_truncate(flags_64);
@@ -1104,18 +1058,17 @@ impl<F: FileSystem + Sync> Server<F> {
         let parent = in_header.nodeid.into();
         let name = dir_entry.name.to_bytes();
         let entry = if name == CURRENT_DIR_CSTR || name == PARENT_DIR_CSTR {
-            // Don't do lookups on the current directory or the parent directory. Safe because
-            // this only contains integer fields and any value is valid.
-            let mut attr = unsafe { MaybeUninit::<libc::stat64>::zeroed().assume_init() };
-            attr.st_ino = dir_entry.ino;
-            attr.st_mode = dir_entry.type_ << 12;
-
             // We use 0 for the inode value to indicate a negative entry.
             Entry {
                 inode: 0,
                 generation: 0,
-                attr,
-                attr_flags: 0,
+                // Don't do lookups on the current directory or the parent directory, i.e. leave
+                // most fields 0.
+                attr: Attr {
+                    ino: dir_entry.ino,
+                    mode: dir_entry.type_ << 12,
+                    ..Default::default()
+                },
                 attr_timeout: Duration::from_secs(0),
                 entry_timeout: Duration::from_secs(0),
             }
@@ -1164,11 +1117,13 @@ impl<F: FileSystem + Sync> Server<F> {
                 let mut err = None;
                 while let Some(dirent) = entries.next() {
                     let mut entry_inode = None;
-                    match self.handle_dirent(&in_header, dirent).and_then(|(d, e)| {
-                        entry_inode = Some(e.inode);
-                        let remaining = (size as usize).saturating_sub(total_written);
-                        add_dirent(&mut cursor, remaining, d, Some(e))
-                    }) {
+                    let bytes_written =
+                        self.handle_dirent(&in_header, dirent).and_then(|(d, e)| {
+                            entry_inode = Some(e.inode);
+                            let remaining = (size as usize).saturating_sub(total_written);
+                            add_dirent(&mut cursor, remaining, d, Some(e))
+                        });
+                    match bytes_written {
                         Ok(0) => {
                             // No more space left in the buffer but we need to undo the lookup
                             // that created the Entry or we will end up with mismatched lookup
@@ -1318,15 +1273,7 @@ impl<F: FileSystem + Sync> Server<F> {
             extensions,
         ) {
             Ok((entry, handle, opts)) => {
-                let entry_out = EntryOut {
-                    nodeid: entry.inode,
-                    generation: entry.generation,
-                    entry_valid: entry.entry_timeout.as_secs(),
-                    attr_valid: entry.attr_timeout.as_secs(),
-                    entry_valid_nsec: entry.entry_timeout.subsec_nanos(),
-                    attr_valid_nsec: entry.attr_timeout.subsec_nanos(),
-                    attr: Attr::with_flags(entry.attr, entry.attr_flags),
-                };
+                let entry_out = EntryOut::from(entry);
                 let open_out = OpenOut {
                     fh: handle.map(Into::into).unwrap_or(0),
                     open_flags: opts.bits(),
@@ -1357,7 +1304,8 @@ impl<F: FileSystem + Sync> Server<F> {
         }
     }
 
-    fn destroy(&self) -> usize {
+    /// Public so the `VhostUserBackend` implementation can call this on device reset
+    pub fn destroy(&self) -> usize {
         // No reply to this function.
         self.fs.destroy();
         self.options
@@ -1522,6 +1470,20 @@ impl<F: FileSystem + Sync> Server<F> {
     }
 }
 
+impl<F: FileSystem + SerializableFileSystem + Sync> SerializableFileSystem for Server<F> {
+    fn prepare_serialization(&self, cancel: Arc<AtomicBool>) {
+        self.fs.prepare_serialization(cancel)
+    }
+
+    fn serialize(&self, state_pipe: File) -> io::Result<()> {
+        self.fs.serialize(state_pipe)
+    }
+
+    fn deserialize_and_apply(&self, state_pipe: File) -> io::Result<()> {
+        self.fs.deserialize_and_apply(state_pipe)
+    }
+}
+
 fn reply_readdir(len: usize, unique: u64, mut w: Writer) -> Result<usize> {
     let out = OutHeader {
         len: (size_of::<OutHeader>() + len) as u32,
@@ -1728,7 +1690,7 @@ fn parse_security_context(nr_secctx: u32, data: &[u8]) -> Result<Option<SecConte
     Ok(Some(fuse_secctx))
 }
 
-fn parse_sup_groups(data: &[u8]) -> Result<u32> {
+fn parse_sup_groups(data: &[u8]) -> Result<GuestGid> {
     let (group_header, group_id_bytes) = take_object::<SuppGroups>(data)?;
 
     // The FUSE extension allows sending several group IDs, but currently the guest
@@ -1738,7 +1700,7 @@ fn parse_sup_groups(data: &[u8]) -> Result<u32> {
     }
 
     let (gid, _) = take_object::<u32>(group_id_bytes)?;
-    Ok(gid)
+    Ok(gid.into())
 }
 
 fn get_extensions(options: FsOptions, skip: usize, request_bytes: &[u8]) -> Result<Extensions> {
